/***************************************************************************
                          pngloader.cpp  -  description
                             -------------------
    begin                : Thu May 1 2003
    copyright            : (C) 2003 by Wojciech Jarosz
    email                : wkjarosz@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "pngloader.h"
#include "byteraster.h"
#include <vector>

pngLoader::pngLoader() :
	imageLoader(),
	number_of_passes(1)
{

}

pngLoader::~pngLoader()
{
}

bool pngLoader::read_image(const char * filename)
{
	infile = fopen(filename, "rb");
	if(!infile)
	{
		std::cerr << "pngLoader: Cannot open file \"" << filename << "\"" << std::endl;
		delete img;
		img = NULL;
		return false;
	}

	// The last three arguments can be used to set up error handling callbacks.
	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if(!png_ptr)
	{
		fclose(infile);
		std::cerr << "pngLoader: Cannot initialize png reader" << std::endl;
		delete img;
		img = NULL;
		return false;
	}

	// Allocate required structure to hold memory information.
	info_ptr = png_create_info_struct(png_ptr);
	if(!info_ptr)
	{
		fclose(infile);
		png_destroy_read_struct(&png_ptr, NULL, NULL);
		std::cerr << "pngLoader: Cannot initialize png reader" << std::endl;
		delete img;
		img = NULL;
		return false;
	}

	// Because we didn't set up any error handlers, we need to be
	// prepared to handle longjmps out of the library on error
	// conditions.
	if(setjmp(png_ptr->jmpbuf))
	{
		png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
		fclose(infile);
		std::cerr << "pngLoader: Reading file failed" << std::endl;
		delete img;
		img = NULL;
		return false;
	}

	png_init_io(png_ptr, infile);

	// Read in all the image information
	png_read_info(png_ptr, info_ptr);

	// Get the header for the first image chunk
	png_uint_32 width, height;
	int bit_depth, color_type, interlace_type;

	png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type, &interlace_type, NULL, NULL);

	////////////////////////////////////////////////////////////////
	// the following tell the PNG library how to transform the image
	// during input

	if(bit_depth == 16)
		// truncate 16 bits/pixel to 8 bits/pixel
		png_set_strip_16(png_ptr);

	if(color_type == PNG_COLOR_TYPE_PALETTE)
		// expand paletted colors into RGB color values
		png_set_expand(png_ptr);
	else if(color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
		// expand grayscale images to full 8 bits/pixel
		png_set_expand(png_ptr);

	if (color_type & PNG_COLOR_MASK_ALPHA)
		png_set_strip_alpha(png_ptr);

	// Expand paletted or RGB images with transparency to full alpha
	// channels so the data will be available as RGBA quartets.
	if(png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
		png_set_expand(png_ptr);

	// update the palette and info structure
	png_read_update_info(png_ptr, info_ptr);

	// read the image data
	std::vector<png_bytep> row_pointers(height);
	int row;
	int nchan = png_get_channels(png_ptr, info_ptr);
	int nbytes = png_get_rowbytes(png_ptr, info_ptr);

	for (row = 0; row < (int)height; row++)
		row_pointers[row] = (png_bytep)malloc(nbytes);

	png_read_image(png_ptr, &row_pointers.front());
	png_read_end(png_ptr, info_ptr);

	// Read it into a ByteRaster structure
	delete img;
	img = new ByteRaster(width, height, nchan);

	unsigned char *pixel = img->pixel(0,0);
	for(row = 0; row < (int)height; row++)
	{
		memcpy(pixel, row_pointers[row], nbytes);
		pixel += nbytes;
	}

	png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
	for(row = 0; row < (int)height; row++)
		free(row_pointers[row]);

	fclose(infile);
	return true;
}

bool pngLoader::start_incremental_image_read(const char * filename)
{
	if (reading)
		return false;
	reading = true;
	
	infile = fopen(filename, "rb");
	if(!infile)
	{
		std::cerr << "pngLoader: Cannot open file \"" << filename << "\"" << std::endl;
		delete img;
		img = NULL;
		return false;
	}

	// The last three arguments can be used to set up error handling callbacks.
	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if(!png_ptr)
	{
		fclose(infile);
		std::cerr << "pngLoader: Cannot initialize png reader" << std::endl;
		delete img;
		img = NULL;
		return false;
	}

	// Allocate required structure to hold memory information.
	info_ptr = png_create_info_struct(png_ptr);
	if(!info_ptr)
	{
		fclose(infile);
		png_destroy_read_struct(&png_ptr, NULL, NULL);
		std::cerr << "pngLoader: Cannot initialize png reader" << std::endl;
		delete img;
		img = NULL;
		return false;
	}

	// Because we didn't set up any error handlers, we need to be
	// prepared to handle longjmps out of the library on error
	// conditions.
	if(setjmp(png_ptr->jmpbuf))
	{
		png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
		fclose(infile);
		std::cerr << "pngLoader: Reading file failed" << std::endl;
		delete img;
		img = NULL;
		return false;
	}

	png_init_io(png_ptr, infile);

	// Read in all the image information
	png_read_info(png_ptr, info_ptr);

	// Get the header for the first image chunk
	png_uint_32 width, height;
	int bit_depth, color_type, interlace_type;

	png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type, &interlace_type, NULL, NULL);

	if (interlace_type == PNG_INTERLACE_ADAM7)
		number_of_passes = png_set_interlace_handling(png_ptr);

	////////////////////////////////////////////////////////////////
	// the following tell the PNG library how to transform the image
	// during input

	if(bit_depth == 16)
		// truncate 16 bits/pixel to 8 bits/pixel
		png_set_strip_16(png_ptr);

	if(color_type == PNG_COLOR_TYPE_PALETTE)
		// expand paletted colors into RGB color values
		png_set_expand(png_ptr);
	else if(color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
		// expand grayscale images to full 8 bits/pixel
		png_set_expand(png_ptr);

	// Expand paletted or RGB images with transparency to full alpha
	// channels so the data will be available as RGBA quartets.
	if(png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
		png_set_expand(png_ptr);

	if (color_type & PNG_COLOR_MASK_ALPHA)
		png_set_strip_alpha(png_ptr);

	// update the palette and info structure
	png_read_update_info(png_ptr, info_ptr);

	int nchan = png_get_channels(png_ptr, info_ptr);

	// Create a ByteRaster structure
	delete img;
	img = new ByteRaster(width, height, nchan);

	scanline = img->head();
	row_stride = width * nchan;
	
	return true;
}

int pngLoader::incremental_image_read(const int& numLines)
{
	int i;
	int num_times = numLines*number_of_passes;
	for (i = 0; i < num_times; i++)
	{
		png_read_row(png_ptr, scanline, NULL);
		scanline += row_stride;
		if (scanline >= img->head() + row_stride*img->height() || i >= img->height())
		{
			scanline = img->head();
		}
	}

	return i/number_of_passes;
}

void pngLoader::finish_incremental_image_read(void)
{
	png_read_end(png_ptr, info_ptr);

	png_destroy_read_struct(&png_ptr, &info_ptr, NULL);

	fclose(infile);
}
