/*!	\file utility.h

	Copyright (C) 2002 Wojciech Jarosz

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef UTILITY_INCLUDED
#define UTILITY_INCLUDED

#include "vec3.h"
#include <math.h>
#include "random.h"

//!	Calculates the ArcCosine.
inline double ArcCos(double x)
{
	double y;

	if (-1.0 <= x && x <= 1.0)
		y = acos( x );
	else if (x >  1.0)
		y = 0.0;
	else if (x < -1.0)
		y = M_PI;

	return y;
}

inline double Sqr(const double &x)
{
	return x*x;
}

//!	Returns a random number.
/*!
	This function returns a uniform pseudo-random double in the half open interval of [0,1).
	The method used is the Mersenne Twister random number generator.
*/
inline double rB_random()
{
	return (double)genrand() * 2.3283064365386963e-10;
}


//!	Returns a signed random number.
/*!
	This function returns a uniform pseudo-random double in the half open interval of [-1,1).
	The method used is the Mersenne Twister random number generator.
*/
inline double signedrandom()
{
	return 2*rB_random()-1;
}

//!	Returns a random byte.
/*!
	This function returns a uniform pseudo-random char in the interval of [0,255].
	The method used is the Mersenne Twister random number generator.
*/
inline char randomByte()
{
	return (char)(genrand() & 0xff);
}

//!	Returns a random double with a normal distribution.
/*!
	This function returns a pseudo-random double in the interval of (-infinity,infinity)
	with a normal distribution.

	The method used is the Mersenne Twister random number generator.
	\param m The mean of the normal distribution.
	\param s The standard deviation of the normal distribution.
	\return A random number with a normal distribution defined by \a m and \a s.
*/
inline double normalrandom(double m, double s)
{
	// normal distribution with mean m and standard deviation s
	double x1, x2, w;
	do
	{
		x1 = 2. * rB_random() - 1.;
		x2 = 2. * rB_random() - 1.;
		w = x1*x1 + x2*x2;
	}
	while (w >= 1. || w < 1E-30);

	w = sqrt((-2.*log(w))/w);

	x1 *= w;

	// x2 *= w;  // a second normally distributed result not used
	return x1 * s + m;
}

//!	Returns a modulus b.
inline double mod(double a, double b)
{
	int n = (int)(a/b);
	a -= n*b;
	if (a < 0)
		a += b;
	return a;
}

//!	Returns a double representing the sign of \a x.
/*!
	\return -1.0 if \a x is negative, and +1.0 if \a x is positive.
*/
inline double sign(double x)
{
	return x > 0. ? 1 : (x < 0. ? -1 : 0);
}

//!	Returns the square root of a number.
/*!
	\return The square root of a number, if it is greater than or equal
	to 0, otherwise, 0.
*/
inline double Sqrt(double x)
{
	return x > 0.0 ? sqrt(x) : 0.0;
}

//!	Returns the integral of t<sup>2</sup>
inline double integral(double t)
{
	return sign(t) * 0.5 * t*t;
}

//!	A filtered absolute value function.
/*!
	This is a smoothed out abs function that preserves C1 continuity.
*/
inline double filteredabs(double x, double dx)
{
	double x0 = x - 0.5*dx;
	double x1 = x0 + dx;
	return (integral(x1)-integral(x0))/dx;
}

//!	Vector valued absolute value function
/*!
	\return The component-wise absolute value of a vector.
*/
inline Vec3 fabs(const Vec3& a)
{
	return Vec3(fabs(a[0]),fabs(a[1]),fabs(a[2]));
}

//!	Linear interpolation.
/*!
	Linearly interpolates between \a a and \a b, using parameter t.
	\param a A value.
	\param b Another value.
	\param t A value between 0 and 1.
	\return A value between a and b based on \a t.
*/
inline double lerp(const double& a, const double& b, const double& t)
{
	if (t <= 0)
		return a;
	else if (t >= 1)
		return b;
	return  a*(1-t) + b*t;
}

//!	Linear interpolation.
/*!
	Linearly interpolates between \a a and \a b, using parameter t.
	\param a A vector (color).
	\param b Another vector.
	\param t A value between 0 and 1.
	\return A vector between a and b based on \a t.
*/
inline Vec3 lerp(const Vec3& a, const Vec3& b, const double& t)
{
	if (t <= 0)
		return a;
	else if (t >= 1)
		return b;
	return  a*(1-t) + b*t;
}

//! Smoothly interpolates between a and b.
/*!
	Does a smooth s-curve (Hermite) interpolation between two values.
	\param	a A value.
	\param	b Another value.
	\param	x A number between \a a and \a b.
*/
inline double smoothstep(const double& a, const double& b, const double& x)
{
	if (x <= a)
		return 0;
	else if (x >= b)
		return 1;
	double temp = (x - a) / (b - a);
	return temp*temp*(3-2*temp);
}

//! Smoothly interpolates between a and b.
/*!
	Does a smooth s-curve (Hermite) interpolation between two vectors (component-wise).
	\param	a A vector.
	\param	b Another vector.
	\param	x A vector of values between \a a and \a b.
*/
inline Vec3 smoothstep(const Vec3& a, const Vec3& b, const Vec3& x)
{
	return Vec3(smoothstep(a[0],b[0],x[0]),smoothstep(a[1],b[1],x[1]),smoothstep(a[2],b[2],x[2]));
}

//! Clamps a double between two bounds.
/*!
	\param	x The value to test.
	\param	a The lower bound.
	\param	b The upper bound.
	\return The value x clamped to the lower and upper bounds.
*/
inline double clamp(const double& x, const double& a, const double& b)
{
	return (x < a ? a : (x > b ? b : x));
}


//! Clamps a int between two bounds.
/*!
	\param	x The value to test.
	\param	a The lower bound.
	\param	b The upper bound.
	\return The value x clamped to the lower and upper bounds.
*/
inline double clamp(const int& x, const int& a, const int& b)
{
	return (x < a ? a : (x > b ? b : x));
}

//! Clamps a vector between two bounds.
/*!
	\param	x The vector to test.
	\param	a The lower bound.
	\param	b The upper bound.
	\return The vector x component-wise clamped to the lower and upper bounds.
*/
inline Vec3 clamp(const Vec3& x, const Vec3& a, const Vec3& b)
{
	return Vec3(clamp(x[0],a[0],b[0]),clamp(x[1],a[1],b[1]),clamp(x[2],a[2],b[2]));
}

//! Adapted version of Perlin's Gamma function.
inline double Gamma(double x, double y)
{
	return pow(x,1.0/y);
}

//! Adapted version of Perlin's Gamma function for vectors (colors).
inline Vec3 Gamma(const Vec3& x, const Vec3& y)
{

	return Vec3(Gamma(x[0],y[0]),Gamma(x[1],y[1]),Gamma(x[2],y[2]));
}

//! Adapted version of Perlin's Bias function.
inline double Bias(const double& x, const double& a)
{
	return x/((1.0/a - 2) * (1-x) + 1);
}

//! Adapted version of Perlin's Bias function for vectors (colors).
inline Vec3 Bias(const Vec3& x, const Vec3& a)
{
	return Vec3(Bias(x[0],a[0]),Bias(x[1],a[1]),Bias(x[1],a[1]));
}

//! Adapted version of Perlin's Gain function.
inline double Gain(const double& x, const double& a)
{
	if (x < 0.5)
	{
		return x/((1.0/a - 2) * (1-2*x) + 1);
	}
	else
	{
		return ((1.0/a - 2) * (1-2*x) - x)/((1.0/a - 2) * (1-2*x) - 1);
	}
}

//! Adapted version of Perlin's Gain function for vectors (colors).
inline Vec3 Gain(const Vec3& x, const Vec3& a)
{
	return Vec3(Gain(x[0],a[0]),Gain(x[1],a[1]),Gain(x[1],a[1]));
}

//! Creates a smooth "spike" function
inline double smoothpulse(const double& e0, const double& e1, const double& e2, const double& e3, const double& x)
{
	return smoothstep(e0,e1,x) - smoothstep(e2,e3,x);
}

//! A repetition of smoothpulses
inline double smoothpulsetrain(const double& e0, const double& e1, const double& e2, const double& e3, const double& period, const double& x)
{
	return smoothpulse(e0,e1,e2,e3,mod(x,period));
}


//! Cosine interpolates between two values.
/*!
	\param	a A value to interpolate.
	\param	b The other value to interpolate between.
	\param	x A number between 0 and 1.
	\return A value between \a a and \a b cosine interpolated using \a x.
*/
inline double Cosine_Interpolate(const double& a, const double& b, const double& x)
{
	double ft = x * M_PI;
	double f = (1 - cos(ft)) * .5;

	return  a*(1-f) + b*f;
}

//! Cosine interpolates between two vectors.
/*!
	\param	a A vector to interpolate.
	\param	b The other vector to interpolate between.
	\param	x A number between 0 and 1.
	\return A vector (color) between \a a and \a b cosine interpolated using \a x.
*/
inline Vec3 Cosine_Interpolate(const Vec3& a, const Vec3& b, const double& x)
{
	double ft = x * M_PI;
	double f = (1 - cos(ft)) * .5;

	return  a*(1-f) + b*f;
}

//! Create an random ortho-normal basis from a vector.
/*!
	Given a vector N, it creates two vectors, X and Y which together
	form an orthogonal basis. ie: they are all mutially orthogonal.

	This algorithm pickes the minor axis in order to reduce numerical
	instability.

	\param	N The vector to create the ortho-normal basis from.
	\param	X Stores the calculated X axis.
	\param	Y Stores the calculated Y axis.
*/
inline void ortho_basis(const Vec3& N, Vec3& X, Vec3& Y)
{
	double min = FEQ_INF;
	int min_index = -1;

	// find the minor axis of the ray
	for (int i = 0; i < 3; i++)
	{
		if (fabs(N[i]) < min)
		{
			min = fabs(N[i]);
			min_index = i;
		}
	}

	if (min_index == 0)
	{
		X = Unit(Vec3(0,-N[2],N[1]));
		Y = Unit(cross(X,N));
	}
	else if (min_index == 1)
	{
		X = Unit(Vec3(-N[2],0,N[0]));
		Y = Unit(cross(X,N));
	}
	else if (min_index == 2)
	{
		X = Unit(Vec3(-N[1],N[0],0));
		Y = Unit(cross(X,N));
	}
}

//! Create an random vector in a unit sphere.
/*!
	This function uses rejection sampling.
	\return A random point in space within a unit sphere.
*/
inline Vec3 spray()
{
	Vec3 xyz;

	do
	{
		xyz[0] = signedrandom();
		xyz[1] = signedrandom();
		xyz[2] = signedrandom();

	} while (norm2(xyz) > 1);


	unitize(xyz);

	return xyz;
}

//! Create an random vector on a hemisphere.
/*!
	\param N Normal of the hemisphere.
	\return A random point in the unit hemisphere defined by N.
*/
inline Vec3 spray(const Vec3& N)
{
	Vec3 X, Y;
	ortho_basis(N,X,Y);
	double theta	= acos(1.0-2.0*rB_random());
	double phi		= 2.0*M_PI*rB_random();

	return X*cos(phi)*sin(theta) + Y*sin(phi)*sin(theta) + N*cos(theta);
}

//! Test whether a vector is very small.
/*!
	\return True if each component of the vector is very small. ie. The vector's length must be small too.
*/
inline bool IsZeroVec(const Vec3& A)
{
	return (IsZero(A[0]) && IsZero(A[1]) && IsZero(A[2]));
}

//! Test whether a vector is not very small.
/*!
	\return True if none of the components of the vector are very small. ie. The vector's length must not be small.
*/
inline bool NonZeroVec(const Vec3& A)
{
	return (!IsZero(A[0]) || !IsZero(A[1]) || !IsZero(A[2]));
}

//! Get the average value of a vector.
/*!
	\return The component-wise average of the vector.
*/
inline double averageVec(Vec3 A)
{
	return (A[0]+A[1]+A[2])/3.0;
}

//! Transforms the two points from uniform values in [0,1) to a 2D point on a disk.
/*!
	This function will transform the two values so that they represent a point in a unit disk.
	Passing two uniform random values will create a point uniformely distributed on a unit disk.
	The method used is the concentric map.
*/
inline void toUnitDisk(double &r1, double &r2)
{
	double phi, r;
	double a = 2*r1 - 1;
	double b = 2*r2 - 1;

	if (a > -b)
	{
		if (a > b)
		{
			r = a;
			phi = (M_PI/4)*(b/a);
		}
		else
		{
			r = b;
			phi = (M_PI/4)*(2 - a/b);
		}
	}
	else
	{
		if (a < b)
		{
			r = -a;
			phi = (M_PI/4)*(4 + b/a);
		}
		else
		{
			r = -b;
			if (!FEQ(b,0))
				phi = (M_PI/4)*(6 - a/b);
			else
				phi = 0.0;
		}
	}

	r1 = r*cos(phi);
	r2 = r*sin(phi);
}

inline double cubic2(const double &a, const double &b, const double &c, const double &d, const double &s)
{
	double h1 = s*s*(2*s - 3) + 1;		// calculate basis function 1
	double h2 = s*s*(3 - 2*s);			// calculate basis function 2
	double h3 = s*(s*(s-2)+1);			// calculate basis function 3
	double h4 = s*s*(s-1);				// calculate basis function 4

	double T1 = 0.5*(b-a) + 0.5*(c-b);	// the incoming tangent
	double T2 = 0.5*(c-b) + 0.5*(d-c);	// the outgoing tangent

	return h1*b +h2*c + h3*T1 + h4*T2;
}

inline Vec3 cubic2(const Vec3 &a, const Vec3 &b, const Vec3 &c, const Vec3 &d, const Vec3 &s)
{
	return Vec3(cubic2(a[0],b[0],c[0],d[0],s[0]),cubic2(a[1],b[1],c[1],d[1],s[1]),cubic2(a[2],b[2],c[2],d[2],s[2]));
}

// UTILITY_INCLUDED
#endif
