/*!	\file vec3.h

	This file is part of the renderBitch distribution.
	Copyright (C) 2002 Wojciech Jarosz

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	Contact:
		Wojciech Jarosz - renderBitch@renderedRealities.net
		See http://renderedRealities.net/ for more information.
*/

#ifndef VEC3_H
#define VEC3_H

#include "common.h"
#include <math.h>
#include <iostream>

//!	A general 3D vector class.
/*!
	This class handles storing and manipulating 3D vectors.
*/
class Vec3
{
private:
	double data[3];

public:

	//!	Default constructor.
	/*!
		Call this constructor if you do not wish to initialize the vector
		to anything. It is more efficient than the others since it does nothing.
	*/
	inline Vec3() { }

	//!	Parameter constructor.
	/*!
		Initialize the vector with the three parameters.
		\param a Value to set the x component to.
		\param b Value to set the y component to.
		\param c Value to set the z component to.
	*/
	inline Vec3(const double &a, const double &b, const double &c)
	{
		data[0] = a;
		data[1] = b;
		data[2] = c;
	}

	//!	Parameter constructor.
	/*!
		Initialize each component to \a a.
		\param a The value to initialize to.
	*/
	inline Vec3(const double &a)
	{
		data[0] = a;
		data[1] = a;
		data[2] = a;
	}

	//!	Copy constructor.
	/*!
		Initialize the vector to be a copy of \a v.
		\param v The vector to create a copy of.
	*/
	inline Vec3(const Vec3 &v)
	{
		data[0] = v[0];
		data[1] = v[1];
		data[2] = v[2];
	}

	//!	Copy constructor.
	/*!
		Initialize the vector from an array of doubles
		\param v The array of doubles to initialize from.
	*/
	inline Vec3(const double* v)
	{
		data[0] = v[0];
		data[1] = v[1];
		data[2] = v[2];
	}

	//!	Copy constructor.
	/*!
		Initialize the vector from an array of floats
		\param v The array of floats to initialize from.
	*/
	inline Vec3(const float* v)
	{
		data[0] = v[0];
		data[1] = v[1];
		data[2] = v[2];
	}

	//!	Constant casting operator.
	/*!
		Casts this constant Vec3 to an array of doubles as a double pointer.
	*/
	inline operator const double*() const
	{
		return data;
	}

	//!	Casting operator.
	/*!
		Casts this Vec3 to an array of doubles as a double pointer.
	*/
	inline operator double*()
	{
		return data;
	}

	//!	Access operator.
	/*!
		Returns the ith component of the vector.
		\param i The component to return.
		\warning i must be either 0, 1, or 2 in order to get expected results.
	*/
	inline double& operator[](int i)
	{
		return data[i];
	}

	//!	Constant access operator.
	/*!
		Returns the ith component of a constant vector.
		\param i The component to return.
		\warning i must be either 0, 1, or 2 in order to get expected results.
	*/
	inline double operator[](int i) const
	{
		return data[i];
	}

	//!	Assignment operator.
	/*!
		Assigns the values from \a a to this Vec3.
	*/
	inline Vec3& operator=(const Vec3 &a)
	{
		data[0] = a[0];
		data[1] = a[1];
		data[2] = a[2];
		return *this;
	}

	//!	Assignment operator.
	/*!
		Sets all components of this Vec3 to \a a.
	*/
	inline Vec3& operator=(const double &a)
	{
		data[0] = a;
		data[1] = a;
		data[2] = a;
		return *this;
	}

	//!	Scalar multiplication operator.
	/*!
		Multiplies this Vec3 by the scalar value \a a and returns the result.
	*/
	inline Vec3 operator*(const double &s) const
	{
		return Vec3(data[0]*s, data[1]*s, data[2]*s);
	}

	//!	Scalar division operator.
	/*!
		Divides this Vec3 by the scalar value \a a and returns the result.
	*/
	inline Vec3 operator/(const double &s) const
	{
		double inv = 1.0/s;		
		return Vec3(data[0]*inv, data[1]*inv, data[2]*inv);
	}

	//!	Vector subtraction operator.
	/*!
		Subtracts the vector \a a from this Vec3 and returns the result.
	*/
	inline Vec3 operator-(const Vec3 &a) const
	{
		return Vec3(data[0]-a[0], data[1]-a[1], data[2]-a[2]);
	}

	//!	Negation operator.
	/*!
		Negates all components and returns the corresponding Vec3.
	*/
	inline Vec3 operator-() const
	{
		return Vec3(-data[0], -data[1], -data[2]);
	}

	//!	Vector addition operator.
	/*!
		Adds the vector \a a to this Vec3.
	*/
	inline Vec3 operator+(const Vec3 &a) const
	{
		return Vec3(a[0]+data[0], a[1]+data[1], a[2]+data[2]);
	}

	//!	Addition assignment operator.
	/*!
		Sets all components of this Vec3 to \a a.
	*/
	inline Vec3& operator+=(const Vec3 &a)
	{
		data[0] += a[0];
		data[1] += a[1];
		data[2] += a[2];
		return *this;
	}

	//!	Vector subtraction assignment operator.
	/*!
		Component-wise subtracts \a a from this Vec3 and saves.
	*/
	inline Vec3& operator-=(const Vec3 &a)
	{
		data[0] -= a[0];
		data[1] -= a[1];
		data[2] -= a[2];
		return *this;
	}

	//!	Scalar multiplication assignment operator.
	/*!
		Multiplies this Vec3 by the scalar \a a and saves.
	*/
	inline Vec3& operator*=(const double &a)
	{
		data[0] *= a;
		data[1] *= a;
		data[2] *= a;
		return *this;
	}

	//!	Vector multiplication assignment operator.
	/*!
		Component-wise multiplies this Vec3 by the Vec3 \a a and saves.
	*/
	inline Vec3& operator*=(const Vec3 &a)
	{
		data[0] *= a[0];
		data[1] *= a[1];
		data[2] *= a[2];
		return *this;
	}

	//!	Scalar division assignment operator.
	/*!
		Divides this Vec3 by the scalar \a a and saves.
	*/
	inline Vec3& operator/=(const double &a)
	{
		double inv = 1.0/a;
		data[0] *= inv;
		data[1] *= inv;
		data[2] *= inv;
		return *this;
	}

	//!	Vector equivalence operator.
	/*!
		Tests to see if each component of \a is equal to each component of this Vec3.
	*/
	inline bool operator==(const Vec3 &a)
	{
		return (a[0] == data[0] && a[1] == data[1] && a[2] == data[2]);
	}

	//!	Dot product.
	/*!
		Performs the dot product between this Vec3 and \a a, and returns the result.
	*/
	inline double dot(const Vec3 &a) const
	{
		return a[0]*data[0] + a[1]*data[1] + a[2]*data[2];
	}

	//!	Cross product.
	/*!
		Performs the cross product between this Vec3 and \a a, and returns the result.
	*/
	inline Vec3 cross(const Vec3 &a) const
	{
		return Vec3(data[1]*a[2]-data[2]*a[1],
					data[2]*a[0]-data[0]*a[2],
					data[0]*a[1]-data[1]*a[0]);
	}

	//!	Length<sup>2</sup>.
	/*!
		Returns the geometric length<sup>2</sup> of the vector.
	*/
	inline double norm2() const
	{
		return dot(*this);
	}

	//!	Length.
	/*!
		Returns the geometric length of the vector.
	*/
	inline double norm() const
	{
		return sqrt(norm2());
	}

	//!	Normalizes the vector.
	/*!
		Scales each component of the vector in order to get unit
		length without changing direction.

		\return The length of the vector prior to normalization.
	*/
	inline double unitize()
	{
		double l = norm2();
		if(l != 1.0 && l > 0.0)
		{
			l = sqrt(l);
			*this /= l;
		}
		return l;
	}
};

//!	Multiply two Vec3s together.
/*!
	\return The component-wise multiplication of the two Vec3s.
	\relates Vec3
*/
inline Vec3 operator*(const Vec3 &a, const Vec3 &b)
{
	return Vec3(a[0]*b[0], a[1]*b[1], a[2]*b[2]);
}

//!	Divide one Vec3 by another.
/*!
	\return The component-wise division of the two Vec3s.
	\relates Vec3
*/
inline Vec3 operator/(const Vec3 &a, const Vec3 &b)
{
	return Vec3(a[0]/b[0], a[1]/b[1], a[2]/b[2]);
}

//!	Multiply a scalare by a Vec3.
/*!
	\return The a Vec3 with each component multiplied by \a s.
	\relates Vec3
*/
inline Vec3 operator*(const double& s, const Vec3& v)
{
	return Vec3(v[0]*s, v[1]*s, v[2]*s);
}

//!	The dot product of two Vec3s.
/*!
	This is a convenience function that simply performs a.dot(b).
	\return The dot product of \a a and \a b.
	\relates Vec3
*/
inline double dot(const Vec3& a, const Vec3& b)
{
	return a.dot(b);
}

//!	The cross product of two Vec3s.
/*!
	This is a convenience function that simply performs a.cross(b).
	\return The \a a cross \a b.
	\relates Vec3
*/
inline Vec3 cross(const Vec3& a, const Vec3& b)
{
	return a.cross(b);
}

//!	The geometric length<sup>2</sup> of a Vec3.
/*!
	This is a convenience function that simply performs a.norm2().
	\return The length<sup>2</sup> of \a a.
	\relates Vec3
*/
inline double norm2(const Vec3& a)
{
	return a.norm2();
}

//!	The geometric length of a Vec3.
/*!
	This is a convenience function that simply performs a.norm().
	\return The length of \a a.
	\relates Vec3
*/
inline double norm(const Vec3& a)
{
	return a.norm();
}

//!	Normalizes the vector.
/*!
	Scales each component of the vector in order to get unit
	length without changing direction.

	This is a convenience function that simply performs a.unitize().

	\return The length of the vector prior to normalization.
	\relates Vec3
*/
inline double unitize(Vec3& a)
{
    return a.unitize();
}

//!	Remove component parallel to B.
/*!
	This function will remove the component parallel to \a B in \a A.

	\return A copy of \a A but with the component parallel to \a B removed.
  	\sa remCompB(Vec3&, const Vec3&)
	\relates Vec3
*/
inline Vec3 remCompB(const Vec3& A, const Vec3& B)
{
	Vec3 C;
	double x = norm2(B);
	if(x > 0.0)
		C = (A-B) * (dot(A,B) / x);
	else
		C = A;

	return C;
}

//!	Remove component parallel to B.
/*!
	This function will remove the component parallel to \a B in \a A and saves it back into A.

	\sa remCompB(const Vec3&, const Vec3&)
	\relates Vec3
*/
inline void remCompB2(Vec3& A, const Vec3& B)
{
	double x = norm2(B);
	if (x > 0.0)
		A = (A-B) * (dot(A,B) / x);
}

//!	Convert a Vec3 from polar space to cartesian space.
/*!
	Given a Vec3 in spherical coordinates, this function will convert it
	to an equivalent Vec3 in rectangular coordinates.

	\return \a A converted into cartesian coordinates.
	\sa rect2Polar(Vec3&)
	\relates Vec3
*/
inline Vec3 polar2Rect(const Vec3& A)
{
	Vec3 rect;
	rect[0] = A[2]*sin(A[1])*cos(A[0]);
	rect[1] = A[2]*cos(A[1]);
	rect[2] = A[2]*sin(A[1])*sin(A[0]);

	return rect;
}

//!	Convert a Vec3 from cartesian space to polar space.
/*!
	Given a Vec3 in rectangular coordinates, this function will convert it
	to an equivalent Vec3 in spherical coordinates.

	\return \a A converted into spherical coordinates.
	\sa polar2Rect(Vec3&)
	\relates Vec3
*/
inline Vec3 rect2Polar(const Vec3& A)
{
	double p = norm(A);
	double S = sqrt(A[0]*A[0] + A[2]*A[2]);
	Vec3 polar;
	polar[0] = (A[0] > 0) ? asin(A[2]/S): M_PI - asin(A[2]/S);
	polar[1] = acos(A[1]/p);
	polar[2] = p;

	return polar;
}


//!	Returns a Unit length copy of \a A.
/*!
	A convenience function that will Unitize any vector.
	return A Unit length copy of \a A.
	\sa Vec3::unitize(), unitize(Vec3&), Unit(const double&, const double&, const double&)
	\relates Vec3
*/
inline Vec3 Unit(const Vec3& A)
{
	double d = norm2(A);
	return d > 0.0 ? A * sqrt(1.0/d) : Vec3(0.0);
}

//!	Creates a Unit length Vec3 out of \a x, \a y, \a z.
/*!
	A convenience function that will construct a Unit vector in the specified direction.
	\sa Vec3::unitize(), unitize(Vec3&), Unit(const Vec3&)
	\relates Vec3
*/
inline Vec3 Unit(const double& x, const double& y, const double& z)
{
	return Unit(Vec3(x,y,z));
}

//!	Output to a stream.
/*!
	Writes \a a to the output stream \a out, and returns the modified stream.
	\return A copy of \a out with the \a a written to it.
	\sa operator>>(std::istream&, Vec3&)
	\relates Vec3
*/
inline std::ostream& operator<<(std::ostream& out, const Vec3& a)
{
	return out << a[0] << " " << a[1] << " " << a[2] ;
}

//!	Input from a stream.
/*!
	Reads \a a from the input stream \a in, and returns the modified stream.
	\return A copy of \a in with the \a a already read from it.
	\sa operator<<(std::ostream&, const Vec3&)
	\relates Vec3
*/
inline std::istream& operator>>(std::istream& in, Vec3& a)
{
	return in >> a[0] >> a[1] >> a[2] ;
}

#endif /* VEC3_H */
